"""Self-play reinforcement learning with actor-critic method
"""
import os
import sys
import numpy as np
import tensorflow as tf
from matplotlib import pyplot as plt
from io import BytesIO
import argparse
import time
import random

import simulator
import baseline

def create_model(height, width, num_output):
  """Create model that should be compatible with 'height' and 'width' as input
  dimensions.

  Args:
    height : desired height for input image
    width  : desired height for input image

  Returns:
    tuple (input, output) where
      input  : placeholder of dtype UINT8 shape [None, height, width, 3] or
               [None, None, None, 3] where we should feed in data
      output : tensor of shape [None, 'num_output'] and dtype FP32

  Note: the first 'None' in 'input' and 'output' shape allows us to potentially
  feed in several states and get output for a whole batch at a time. Even though
  we have to do choose actions sequentially within a game, we could potentially
  play many games simultanously! The second and third 'None' in 'input' also
  allows us to have variable-sized inputs (though all inputs within a batch must
  always have the same height and width). This depends on us being able to cope
  with different image sized though, e.g. by doing some global pooling in our
  model though
  """
  #x = tf.placeholder([None, None, None, 3], dtype=tf.uint8)
  input_tensor = tf.placeholder(dtype=tf.uint8, shape=[None, height, width, 3])
  x = tf.cast(input_tensor, tf.float32) # use float32
  x = (x - 127.5) / 127.5 # scale input to [-1, 1]
  weight_decay = 1e-4

  def layer_template(x, num_filters, kernel_size):
    """Template for conv --> relu layer with some parameters already set"""

    kernel_initializer = tf.variance_scaling_initializer(scale=2.0, distribution="truncated_normal")
    kernel_regularizer = tf.contrib.layers.l2_regularizer(weight_decay)

    x = tf.layers.conv2d(x, num_filters, kernel_size=kernel_size, strides=[1, 1], padding="SAME",
                                kernel_initializer=kernel_initializer,
                                kernel_regularizer=kernel_regularizer
    )

    return x

  x1 = layer_template(x,  num_filters=32, kernel_size=[3, 3])
  x1 = tf.nn.relu(x1)
  x2 = layer_template(x1, num_filters=32, kernel_size=[3, 3])
  x2 = tf.nn.relu(x2)
  x3 = layer_template(x2, num_filters=32, kernel_size=[3, 3]) + x1
  x3 = tf.nn.relu(x3)
  x  = layer_template(x3, num_filters=1 , kernel_size=[1, 1])

  # Take average over all 'height' and 'width' features
  # x = tf.reduce_mean(x, axis=[1, 2], name="global_average_pool")
  x = tf.layers.flatten(x)
  kernel_initializer = tf.zeros_initializer()
  kernel_regularizer = tf.contrib.layers.l2_regularizer(weight_decay)
  output = tf.layers.dense(x, num_output, kernel_initializer=kernel_initializer, kernel_regularizer=kernel_regularizer)

  return input_tensor, output

class ValueNetwork:
  """Value network to predict value of a state for the policy"""

  def __init__(self, height, width, sess, learning_rate=None):
    """
    Args:
      height        : height of board
      width         : width of board
      sess          : session that can be used to run operations
      learning_rate : step length for parameter update used during training
    """

    self.learning_rate = learning_rate
    self.sess = sess
    self.value_op = self._value_op(height, width)
    self.target_tensor = tf.placeholder(dtype=tf.float32, shape=[None])
    if self.learning_rate:
      self.update_op = self._update_op(self.target_tensor)

  def _value_op(self, height, width):
    """Creates value network

    Args:
      height  : height of board
      width   : width of board

    Returns:
      A tensor that can be run to obtain value estimate of a state. We need to
      feed in a state using 'feed_dict' argument with 'self.input_tensor' as key
    """

    with tf.name_scope("value"):
      input_tensor, preactivation = create_model(height, width, num_output=1)
      # use cross-entropy or squared error loss?
      self.input_tensor = input_tensor
      value = tf.tanh(preactivation)

    return value

  def _update_op(self, target):
    """Creates op that run update of parameters by calculating gradient of

      (value(state) - target)^2

    with respect to parameters, and applying an update in the opposite direction
    of the gradient, with step length equal to self.learning_rate.

    Args:
      Tensor of type FP32 and shape [batch_size]
    """

    with tf.name_scope("grad"):
      # we will use the optimizer to calulate gradients
      optimizer = tf.train.GradientDescentOptimizer(learning_rate=self.learning_rate)
      self.loss = tf.reduce_mean((self.value_op - target)**2)
      
      grads_and_vars = optimizer.compute_gradients(self.loss)
      # here we could potentially alter gradient in some way
      apply_grads = optimizer.apply_gradients(grads_and_vars)

    return apply_grads

  def value(self, state):
    """Calculate value(state)

    Args:
      state : numpy array of dimensions
              - [batch_size x height x width x 3], or
              - [height x width x 3]
    Returns:
      float values in [-1, 1] with value estimates for each state, shape is either
      - numpy array of [batch_size] if input has batch dimension
      - otherwise a scalar float value
    """

    is_batch = len(state.shape) == 4
    if not is_batch:
      # [height x width x 3] --> [1, height x width x 3]
      state = np.expand_dims(state, axis=0)

    v = self.sess.run(self.value_op, feed_dict={self.input_tensor : state})
    if not is_batch:
      # shape [1] --> scalar
      v = float(v)

    return v

  def update_params(self, state, target):
    """Update parameters by formula

       params += learning_rate*(target - value(state))grad_params(value(state))
    
    Args:
      state : numpy array of dimensions
              - [batch_size x height x width x 3], or
              - [height x width x 3]
      target : targets for value(state), either,
              - numpy array with shape [batch_size] with targets for each shape, or
              - scalar float target for single state

    Returns:
      loss : float, mean((value-target)^2), only returned for summary purposes
    """

    is_batch = len(state.shape) == 4
    if not is_batch:
      # [height x width x 3] --> [1, height x width x 3]
      state = np.expand_dims(state, axis=0)
    if np.isscalar(target):
      target = [target]

    loss, _, = self.sess.run([self.loss, self.update_op], feed_dict={
      self.input_tensor  : state,
      self.target_tensor : target
    })
    return loss

  def _grad_op_value(self):
    """Calculate and return two ops
    1. to get Gradient of value(state) with respect to parameters
    2. to apply gradients

    Note that this is not currently in use, but could be used for a more
    efficient implementation. In our current approach we are calculating
    value(state) in one operation, and then later we calculate the gradient with
    respect to (target - value(state))^2, where we feed in the target. This is
    not very efficient as we do the forward pass, i.e. value(state) twice, as we
    redo it when we calculate the gradient. An alternative to this is to
    calculate the gradient of of value(state) with respect to the parameters
    right away. When we get the target at the next operation, value(next_state)
    + r, we can alter the computed gradients (which we have stored in numpy
    arrays) by applying the scalar (target-value) to them, and then feed this in
    to the apply_gradients operation. This does require some more effort from
    our part though.
    """

    with tf.name_scope("grad_value"):
      # we will use the optimizer to calulate gradients
      optimizer = tf.train.GradientDescentOptimizer(learning_rate=self.learning_rate)
      
      grads_and_vars = optimizer.compute_gradients(self.value)
      grads = [grad for grad, var in grads_and_vars]
      grad_dict = {grad.name : grad for grad in grads}

      apply_grads = optimizer.apply_gradients(grads_and_vars)

    return grad_dict, apply_grads

# TODO: implement policy network
# This network runs as it is, but no training happens
class PolicyNetwork:

  def __init__(self, height, width, sess):

    self.v = tf.Variable(0.0, name="variable") # you should remove this
    self.random_policy = baseline.SmartRandomPolicy() # you should remove this

  def next_action(self, state):

    # you should remove this
    to_int = {
      "left"  : 0,
      "right" : 1,
      "up"    : 2,
      "down"  : 3
    }
    # you should remove this
    return to_int[self.random_policy.next_action(state)]

def initialize_networks(sess, ckptdirs, height, width, learning_rate):
  """Initialize policy networks for agent and opponent, and value network.

  Args:
    sess             :
    ckptdirs         : dict with paths to checkpoint-directories
      agent             : ckpt-dir for agent-network
      opponent          : ckpt-dir for opponent-network
      value_network     : ckpt-dir for value-network
    height           : height of board
    width            : width of board

  Returns:
    agent            : policy network with weights for agent
    opponent         : policy network with opponent-weights
    value_network    : value network
    savers           : dict with savers
      agent            : saver for agent
      opponentr        : saver for opponent
      value_network    : saver for value network
  """

  with tf.variable_scope("agent"):
    # TODO: this will probably change depndant on your implementation on policy network...
    agent = PolicyNetwork(height, width, sess)
  agent_saver = tf.train.Saver(max_to_keep=5, keep_checkpoint_every_n_hours=2)

  with tf.variable_scope("opponent"):
    opponent = PolicyNetwork(height, width, sess)
  opponent_vars = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, scope="opponent")
  # variables in the checkpoint are stored as /agent/var_name. We want to map
  # theses varaiable names to the corresponding variables of the opponent, then
  # in the 'opponent_saver.restore' call, the weights from the checkpoint will
  # be loaded into the correct variable of 'opponent'
  # opponent/var_name:0 --> agent/var_name
  ckptname_to_opponent_var = {v.name.replace("opponent/", "agent/", 1).replace(":0", "") : v for v in opponent_vars}
  opponent_saver = tf.train.Saver(ckptname_to_opponent_var, max_to_keep=5)

  with tf.variable_scope("value_network"):
    value_network = ValueNetwork(height, width, sess, learning_rate=learning_rate)
  value_vars = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, scope="value_network")
  value_saver = tf.train.Saver(value_vars, max_to_keep=5)

  init_op = tf.global_variables_initializer()

  sess.run(init_op)

  # if no checkpoints --> create checkpoint...
  latest_checkpoint = tf.train.latest_checkpoint(ckptdirs["agent"])
  if latest_checkpoint is None:
    # create checkpoint
    agent_saver.save(sess, os.path.join(ckptdirs["agent"], "model.ckpt"), global_step=0)
    agent_saver.save(sess, os.path.join(ckptdirs["opponent"], "model.ckpt"), global_step=0)
    latest_checkpoint = tf.train.latest_checkpoint(ckptdirs["agent"])
  else:
    agent_saver.restore(sess, latest_checkpoint)

  opponent_checkpoint = tf.train.latest_checkpoint(ckptdirs["opponent"])
  opponent_saver.restore(sess, opponent_checkpoint)

  # restore value function parameters
  value_checkpoint = tf.train.latest_checkpoint(ckptdirs["value_network"])
  if value_checkpoint:
    value_saver.restore(sess, value_checkpoint)

  savers = {
    "agent"         : agent_saver,
    "opponent"      : opponent_saver,
    "value_network" : value_saver
  }

  return agent, opponent, value_network, savers

# TODO: need to train agent
def train(agent,
          environment,
          value_network,
          savers,
          ckptnames,
          win_threshold,
          learning_rate,
          summary_writer,
          num_games,
          inc_num_games,
          sess):
  """Train 'agent' in 'environment' with actor-critic method

  If percentage of wins is more than 'win_threshold' we return
  """

  INTERVAL = 500
  plot_value_loss_interval = 500
  CHECKPOINT_INTERVAL = 10000
  num_steps = 0
  start = start_training = time.time()
  mirror_augment = True
  # add initial value prediction from value network?
  num_games_start = num_games
  train_summaries = TrainSummaries()
  while True:

    # randomly pick to go first or second
    player = random.choice(["white", "black"])
    game = environment.new_game(player=player)

    state = game.init_state()
    END_OF_GAME = False
    rewards = [] # really not necessary to have list, as only last reward will be nonzero...

    if mirror_augment:
      old_value_estimate = np.mean(value_network.value(augment_with_mirroring(state)))
    else:
      old_value_estimate = value_network.value(state)

    while not END_OF_GAME:

      action = agent.next_action(state)
      if action < 0 or action > 3:
        raise Exception("Action value '%d' is not valid. Should be in [0, 1, 2, 3]"
                        "You may have NAN values in your graph" % action)

      train_summaries.update_after_move(action)
      num_steps += 1

      silly_loss = not game._surround.is_legal_move(player, action) and game._surround.has_legal_move(player)
      new_state, reward, END_OF_GAME = game.apply_action(action)
      rewards.append(reward)

      if not END_OF_GAME:
        if mirror_augment:
          new_state_augmented = augment_with_mirroring(new_state)
          #figure, axis = plt.subplots(4)
          #for i, ax in enumerate(axis):
          #  ax.imshow(new_state_augmented[i])
          #plt.show()
          # plot std of these values???
          next_value = np.mean(value_network.value(new_state_augmented))
        else:
          next_value = value_network.value(new_state)
        new_value_estimate =  next_value + reward # reward should be 0
      else:
        new_value_estimate = reward
        next_value = 0

      critic = new_value_estimate - old_value_estimate
      old_value_estimate = next_value
      # this does forward propagation again...

      if mirror_augment:
        state = augment_with_mirroring(state)
        new_value_estimate = 4*[new_value_estimate] # list of 4 copies of value

      # update value network
      value_loss = value_network.update_params(state=state, target=new_value_estimate)
      if num_steps % plot_value_loss_interval == 0:
        # note that this loss should not really go to zero.
        # it is only single samples only, so will be very noisy
        add_scalar_summary("value_function_loss", value_loss, num_games, summary_writer)

      state = new_state


    # calculate return, although this will be list of either 0s, 1s or -1s as
    # we only have reward at the end
    # we do not use discouting, i.e. gamma = 1
    T = len(rewards)
    G_t = np.zeros(T, dtype=np.float32)
    G_t[T-1] = rewards[T-1]
    # iterate from end to start
    for t in range(T-1, 0, -1):
      G_t[t-1] = G_t[t] + rewards[t-1]
    assert G_t[0] == G_t[-1], "Should only be reward at end"

    num_games = sess.run(inc_num_games)

    train_summaries.update_after_episode(player, G_t[0], silly_loss, T, state)

    if num_games % INTERVAL == 0:
      games_played = min(INTERVAL, num_games-num_games_start)
      print("num_games: %6d. sec/game: %.3f" % (num_games, (time.time() - start) / games_played))
      win_percentage = train_summaries.add_and_reset_summaries(summary_writer, num_games, learning_rate)
      print("Percentage of last %s games won: '%.3f" % (games_played, win_percentage))

      if win_percentage >= win_threshold and games_played == INTERVAL:
        return num_games, win_percentage

      start = time.time()

    if num_games > 0 and num_games % CHECKPOINT_INTERVAL == 0:
      savers["agent"].save(sess, ckptnames["agent"], global_step=num_games)
      savers["value_network"].save(sess, ckptnames["value_network"], global_step=num_games)

def count_variable(name):
  """Create zero-initalized non-trainable INT32-variable

  Args:
    name : name of variable
  Returns:
    tuple (variable, update_op) where update_op increments count by 1
  """

  count = tf.Variable(tf.zeros([], dtype=tf.int32), trainable=False, name=name)
  inc_count = tf.assign_add(count, 1, name="inc_" + name)

  return count, inc_count

def eval_against_baseline(agent,
                          height,
                          width,
                          summary_writer,
                          num_games,
                          num_eval_games=200
):
  """Evaluate agent against baseline agents

  Args:
    agent  : agent to evaluate, should be object with 'nect_action' methods
    height : height of board
    width  : width of board
    summary_writer : writer to write summaries to TensorBoard event file
    num_games      : number of games used for training
    num_eval_games : how many games to play against each opponent
  """

  b1 = baseline.RandomPolicy()
  b2 = baseline.SmartRandomPolicy()
  win_percentage_b1 = eval_against_opponent(
    agent, b1, num_eval_games, height, width, summary_writer, num_games
  )
  win_percentage_b2 = eval_against_opponent(
    agent, b2, num_eval_games, height, width, summary_writer, num_games,
    num_display_games=9, tag_prefix="eval_random_smart"
  )
  add_scalar_summary(
    "win_p_against_random", win_percentage_b1, num_games, summary_writer
  )
  add_scalar_summary(
    "win_p_against_smart_random", win_percentage_b2, num_games, summary_writer
  )
  print("win_p_against_random", win_percentage_b1)
  print("win_p_against_smart_random", win_percentage_b2)

def eval_against_opponent(agent,
                          eval_opponent,
                          num_eval_games,
                          height,
                          width,
                          summary_writer,
                          num_games,
                          num_display_games=0,
                          tag_prefix=None
):
  """Evaluate 'agent' against 'eval_opponent' for 'num_eval_games'

  Args:
    agent  : agent to evaluate, should be object with 'nect_action' methods
    eval_opponent  : opponent, should be object with 'nect_action' methods
    num_eval_games : how many games to play against each opponent
    height         : height of board
    width          : width of board
    summary_writer : writer to write summaries to TensorBoard event file
    num_games      : number of games used for training
    num_display_games : number of games to plot in TensorBoard
    tag_prefix        : prefix for tag name to use in TensorBoard
  
  Returns:
    win_percentage : percentage of game 'agent' won against 'eval_opponent'
  """

  eval_env = simulator.Environment(height, width, eval_opponent, turn_based=True)

  display_games = []
  num_wins = 0
  for i in range(num_eval_games):
    player = random.choice(["white", "black"])
    game = eval_env.new_game(player=player)
    state = game.init_state()
    END_OF_GAME = False
    while not END_OF_GAME:
      action = agent.next_action(state)
      state, reward, END_OF_GAME = game.apply_action(action)

    # only need to look at last reward
    if reward > 0:
      num_wins += 1

    if i < num_display_games:
      display_games.append(state)

  if num_display_games > 0:
    add_image_summaries(display_games, summary_writer, num_games, tag_prefix=tag_prefix)

  win_percentage = num_wins / num_eval_games

  return win_percentage

def add_image_summaries(images, summary_writer, num_games, tag_prefix=None):
  """Add list of uint8 numpy images to TensorBoard
  
  Args:
    images         : list of [height, width, 3] images with uint8 data
    summary_writer : use to write summary to tensorboard event file
    num_games      : number of games used for training
    tag_prefix     : start of tag name in tensorboard
  """

  if tag_prefix is None:
    tag_prefix = "example_game"

  image_summaries = []
  for i, image in enumerate(images):
    image_data = BytesIO()
    plt.imsave(image_data, image, format='PNG')
    image_proto = tf.Summary.Image(encoded_image_string=image_data.getvalue(),
                             height=image.shape[0], width=image.shape[1])
    image_summaries.append(
      tf.Summary.Value(tag="%s-%d" % (tag_prefix, i), image=image_proto)
    )

  summary = tf.Summary(value=image_summaries)
  summary_writer.add_summary(summary, global_step=num_games)

def add_scalar_summary(tag, value, num_games, summary_writer):
  """Add scalar summary to TensorBoard

  Args:
    tag       : summary name
    value     : scalar value
    num_games : number of games used for training
    summary_writer : use to write summary to tensorboard event file
  """

  summary = tf.Summary(value=[tf.Summary.Value(tag=tag, simple_value=value)])
  summary_writer.add_summary(summary, global_step=num_games)

class TrainSummaries():
  """Summaries we update during training"""

  def __init__(self):

    self.stats = {
      "game_won"     : [],
      "silly_losses" : [],
      "white_result" : [], # results for plays as white player
      "black_result" : [], # results for plays as black playe
      "game_lengths" : []
    }
    # tuples of the form (name_of_summary, stat, postprocess_function)
    self.summaries = [
      ("win_percentage", "game_won", np.mean),
      ("silly_loss_percentage", "silly_losses", np.mean),
      ("win_percentage_white", "white_result", np.mean),
      ("win_percentage_black", "black_result", np.mean),
      ("average_game_lengths", "game_lengths", np.mean)
    ]
    self.move_counts = np.zeros(4)

    self.moves = ["left", "right", "up", "down"]
    self.display_games = 9
    self.idx = 0
    self.last_N_games = self.display_games*[None]

  # action is a bit differn
  def update_after_move(self, action):
    self.move_counts[action] += 1

  def _reset(self):
    """Resest statistics"""

    for key, value in self.stats.items():
      value.clear()

    self.move_counts[:] = 0

  def update_after_episode(self, player, reward, silly_loss, T, state):
    """Update statistics based on episode data

    Args:
      player     : string value holding color of agent, 'white' or 'black'
      reward     : scalar with value 1 if won, -1 if lost
      silly_loss : True if we lost even though we had legal moves to play.
                   False if we lost due to no options left.
                   If we won, the value is undefined and you should disregard it
      T          : time steps for episode
      state      : final state, numpy array of shape [height, width, 3]
    """

    stats = self.stats
    R = reward
    won = R > 0
    lost = not won
    stats["game_won"].append(won)
    if lost:
      stats["silly_losses"].append(silly_loss)
    if player == "white":
      stats["white_result"].append(won)
    else:
      stats["black_result"].append(won)
    stats["game_lengths"].append(T)

    if self.idx >= self.display_games:
      self.idx = 0
    self.last_N_games[self.idx] = state
    self.idx += 1

  def add_and_reset_summaries(self, summary_writer, num_games, learning_rate):
    """Print and add summaires to TensorBoard, then reset statistics """

    stats = self.stats

    for name, stat, f in self.summaries:
      add_scalar_summary(name, f(stats[stat]), num_games, summary_writer)
    win_percentage = np.mean(stats["game_won"])

    move_probs = self.move_counts / np.sum(self.move_counts)
    for move, move_p in zip(self.moves, move_probs):
      add_scalar_summary("p_" + move, move_p, num_games, summary_writer)

    add_scalar_summary("learning_rate", learning_rate, num_games, summary_writer)

    images = [image for image in self.last_N_games if image is not None]
    add_image_summaries(
      images, summary_writer, num_games, tag_prefix="train_game"
    )

    self._reset()

    return win_percentage

def augment_with_mirroring(state, action=None):
  """Create four versions of state (and action) by mirroring
  1. original
  2. up-down mirrored
  3. left-right mirrored
  4. left-right and up-down mirrored

  Args:
    state  : numpy array of dimension [height x width x 3]
    action : int in [0, 1, 2, 3] representing [left, right, up, down] respectively

  Returns:
    either states or (states, actions) tuple of the four mirrored versions
    described above where
      states : numpy array of shape [4, height x width x 3]
      action : numpy array of shape [4]
  """
  states = []
  if action is not None:
    actions = []
  for left_right in [False, True]:
    for up_down in [False, True]:
      r = mirror(state, left_right, up_down, action=action)
      if action is not None:
        s, a = r
        states.append(s)
        actions.append(a)
      else:
        s = r
        states.append(s)

  states = np.array(states)
  if action is None:
    return states

  actions = np.array(actions, dtype=np.int32)

  return states, actions


def mirror(state, left_right, up_down, action=None):
  """Mirror both state and action (if provided) around desired axis

  Args:
    state      : numpy array of shape [height x width x 3]
    action     : int in [0, 1, 2, 3] meaning [left, right, up, down] respectively
    left_right : if True we flip left-right of image and left-right of actions
    up_down    : if True we flip up-down of image and up-down of actions

  Returns:
    tuple (state, action) of mirrored state and action or just mirrored state
    if action is not provided
  """

  if len(state.shape) != 3:
    raise ValueError("Expected shape of the form [height, width, channels]"
                     ", got '%s'" % state.shape)

  if left_right:
    state = state[:, ::-1]
    if action is not None and action in [0, 1]:
      action = 1 if action == 0 else 0
  if up_down:
    state = state[::-1]
    if action is not None and action in [2, 3]:
      action = 3 if action == 2 else 2

  if action is None:
    return state

  return state, action

def main(train_dir, learning_rate):

  height, width = (11, 16)
  sess = tf.Session()

  # keep track of how many games we have played
  num_games_played, inc_num_games = count_variable("num_games")
  # keep track of how many times we have switched opponent
  num_opponent_switches, inc_num_opponents = count_variable("num_opponents")

  ckptdirs = {
    "agent" : train_dir,
    "opponent" : os.path.join(train_dir, "opponent"),
    "value_network" : os.path.join(train_dir, "value_network")
  }
  ckptnames = {name : os.path.join(d, "model.ckpt") for name, d in ckptdirs.items()}

  agent, opponent, value_network, savers = initialize_networks(
    sess, ckptdirs, height, width, learning_rate
  )
  logdir = train_dir
  summary_writer = tf.summary.FileWriter(logdir, flush_secs=30,
                                         graph=tf.get_default_graph())

  environment = simulator.Environment(height, width, opponent, turn_based=True)

  win_threshold = 0.6
  # num_games_played may be nonzero if we are restoring from checkpoint
  num_games = sess.run(num_games_played)
  while True:
    try:
      # for each iteration we train until agent beats opponent with probability
      # at least win_threshold
      num_games, win_percentage = train(
        agent,
        environment,
        value_network,
        savers,
        ckptnames,
        win_threshold,
        learning_rate,
        summary_writer,
        num_games,
        inc_num_games,
        sess
      )

      # evaluate agent against 'random' and 'smart random' baselines
      eval_against_baseline(agent, height, width, summary_writer, num_games)

      # save agent and a copy in "opponent" dir, as well as value network
      savers["agent"].save(sess, ckptnames["agent"], global_step=num_games)
      savers["agent"].save(sess, ckptnames["opponent"], global_step=num_games)
      savers["value_network"].save(sess, ckptnames["value_network"], global_step=num_games)

      # update opponent
      switches = sess.run(inc_num_opponents)
      print("Switching to stronger opponent")
      add_scalar_summary("opponent_switches", switches, num_games, summary_writer)
      savers["opponent"].restore(sess, tf.train.latest_checkpoint(ckptdirs["opponent"]))

    except KeyboardInterrupt:
      print("Saving agent and value network to checkpoint before exiting")
      num_games = sess.run(num_games_played)
      savers["agent"].save(sess, ckptnames["agent"], global_step=num_games)
      savers["value_network"].save(sess, ckptnames["value_network"], global_step=num_games)
      break

  # cleanup
  summary_writer.close()
  sess.close()

def parse_args():
  """Parse arguments form command line"""

  parser = argparse.ArgumentParser()

  parser.add_argument("train_dir", help="Directory to save checkpoint and summaries")
  parser.add_argument("learning_rate", type=float, help="learning rate")

  return parser.parse_args()

if __name__ == '__main__':

  args = parse_args()
  main(args.train_dir, args.learning_rate)
