import random
import numpy as np
from collections import OrderedDict

# TODO: Full support for more than two players, may not want to have more than
# three channels, so to let model take in RGB in any case. Could make all other
# players gray?
class Surround:
  """Multiplayer Surround"""

  def __init__(self, height, width, margin=1, shuffle_init=True):
    """
    Args:
      height       : height of board
      width        : width of board
      margin       : size of margin within board
      shuffle_init : if True shuffles init positions of 'black' and 'white',
                     otherwise 'white' starts to the left, 'black' to the right
    """

    self.height = height
    self.width = width
    self.margin = margin

    h, w = height, width

    init_positions = [(h // 2, w // 3), (h // 2, w-(w// 3)-1)]
    if shuffle_init:
      np.random.shuffle(init_positions)
    init_white, init_black = init_positions

    board = np.zeros((h, w, 2), dtype=np.int32)
    board[init_white] = [1, 0]
    board[init_black] = [0, 1]
    self.board = board

    self.heads = {
      "white" : init_white,
      "black" : init_black
    }
    self.players = ["white", "black"]
    self.index = OrderedDict([(p, i) for i, p in enumerate(self.players)])

    self.moves = OrderedDict([
      ("left",  [ 0, -1]),
      ("right", [ 0,  1]),
      ("up",    [-1,  0]),
      ("down",  [ 1,  0])
    ])

  def next_player(self, player):
    """Figure out which player is next after 'player'"""

    index_player = self.players.index(player)
    index_next_player = (index_player + 1) % len(self.players)
    next_player = self.players[index_next_player]

    return next_player

  def vis_board(self, player):
    """Create board from the point of view of 'player'. We always switch the
    blocks of 'player' to the second (green) axis. In this way the blocks of
    the player will look 'green' for any player from its own perspective. In
    this way we can use the same network to play any of the players.
    """

    # we move the channel of player to the first index
    index = self.index[player]
    channels = self.board.shape[-1]
    l = list(range(channels))
    l.pop(index)
    l.insert(1, index)

    vis_board = np.zeros((self.height, self.width, channels+1), dtype=np.uint8)
    border_image = np.zeros((self.height, self.width, channels+1), dtype=np.uint8)
    vis_board[:, :, :-1] = self.board[:, :, l]

    # insert head positions into last channel to make simpler
    # may then use current frame as state
    for player, head in self.heads.items():
      vis_board[head][-1] = 1

    vis_board *= 255

    for b in range(self.margin):
      border_image[[b, -(b+1)], :, :] = 128
      border_image[:, [b, -(b+1)], :] = 128
    vis_board = np.maximum(vis_board, border_image)

    return vis_board

  def within_play_area(self, pos):
    """Check if 'pos' is on board, and inside border (margin)

    Args:
      pos: tuple of coordinates (y, x)

    Returns:
      True if pos is within board else False
    """
    pos = np.array(pos)

    m = self.margin
    return np.all(pos >= [m, m]) and np.all(pos < [self.height-m, self.width-m])

  def is_occupied(self, pos):
    """Check if 'pos' is occupied by another player or not

    Args:
      pos: coordinates [y, x], should be on board
    
    Returns:
      True if 'pos' is occuppied, otherwise False
    """
    return np.any(self.board[pos] != [0, 0])

  def is_legal_move(self, player, move):
    """
    Args:
      player : one of ['white', 'black']
      move   : one of the valid moves

    Returns:
      True if it is legal, False if not
    """
    move = move if not isinstance(move, int) else list(self.moves)[move]
    next_pos = self.next_pos(player, move)
    is_legal = self.within_play_area(next_pos) and not self.is_occupied(next_pos)

    return is_legal

  def has_legal_move(self, player):
    """Check if a player has any legal moves"""

    has_legal = False
    for move in self.moves:
      if self.is_legal_move(player, move):
        has_legal = True
        break

    return has_legal

  #def check_moves_legal(self, player):
  #  """
  #  Args:
  #    Player: one of ['white', 'black']

  #  Returns:
  #    OrderedDict with moves as keys and value indicating whether move is legal
  #  """
  #  is_legal = OrderedDict()
  #  for move in self.moves:
  #    is_legal[move] = self.is_legal_move(player, move)

  #  return is_legal

  def next_pos(self, player, move):
    """Figure out the next position of the 'player' given that 'move' is applied
    
    Args:
      player : oneof ['white', 'black']
      move   : oneof ['left', 'right', 'up', 'down']
    
    Returns:
      tuple (y, x) of new position
    Note: the returned position may be outside the board!
    """
    current_pos = np.array(self.heads[player])
    update = np.array(self.moves[move])
    new_pos = tuple(current_pos + update)

    return new_pos

  def move_conflict(self, player, moves):
    """Check if 'player' tries to move to the same square as another player

    Args:
      moves: dict with 'player' as key and 'move' as value
    
    Returns:
      True if it tries to move to the same square as another player else False
    """

    conflict = False
    next_pos = self.next_pos(player, moves[player])
    for p, move in moves.items():
      if p == player:
        continue
      if self.next_pos(p, moves[p]) == next_pos:
        conflict = True
        break

    return conflict

  def apply_move_board(self, player, move):
    """Apply move of player to board"""

    old_pos = self.heads[player]
    new_pos = self.next_pos(player, move)

    self.board[new_pos][self.index[player]] = 1 # put down block for player
    self.heads[player] = new_pos # update head position
    
  def apply_actions(self, moves):
    """Apply action for each of the players
    
    Args:
      moves: dict with 'player' as key and 'move' as value.
             'move' may be either int or string

    Returns: 
      tuple with entries
      - boolean which is True if reached terminal state, else False
      - dict with player as key and tuple (next_state, reward). If turn_based
        only these values for the next player will be returned, except if
        reached terminal state, then this will be returned for all players.
    """

    crashes = {}

    ordered_moves = list(self.moves.keys())
    moves = {player : move if not isinstance(move, int) else ordered_moves[move] for player, move in moves.items()}

    for player, move in moves.items():
      if isinstance(move, int):
        move = ordered_moves[move]
        player[move] = move
      pos = self.next_pos(player, move)
      crashes[player] = not self.is_legal_move(player, move) or self.move_conflict(player, moves)
      self.apply_move_board(player, move)

    players = list(moves.keys())
    if len(moves) == 1:
      # assumes turn-based
      player = players[0]
      next_player = self.next_player(player)

      if crashes[player]:
        # game has ended
        END_OF_GAME = True
        state_reward = {
          player      : (self.vis_board(player), -1),
          next_player : (self.vis_board(next_player), 1)
        }
      else:
        END_OF_GAME = False
        board = self.vis_board(next_player)
        state_reward = {
          next_player : (board, 0)
        }
      return state_reward, END_OF_GAME

    elif len(moves) != 2:
      raise NotImplementedError("Can only handle two players for now")

    p1, p2 = crashes.keys()

    state_reward = {}
    END_OF_GAME = True
    p1_board = self.vis_board(p1)
    p2_board = self.vis_board(p2)
    if crashes[p1] and crashes[p2]:
      # draw
      state_reward[p1] = (p1_board, 0)
      state_reward[p2] = (p2_board, 0)
    elif not crashes[p1] and crashes[p2]:
      # player 1 wins
      state_reward[p1] = (p1_board, 1)
      state_reward[p2] = (p2_board, -1)
    elif crashes[p1] and not crashes[p2]:
      # player 2 wins
      state_reward[p1] = (p1_board, -1)
      state_reward[p2] = (p2_board, 1)
    else:
      # keep on playing
      END_OF_GAME = False
      state_reward[p1] = (p1_board, 0)
      state_reward[p2] = (p2_board, 0)

    return state_reward, END_OF_GAME

# Do serially for now...?
class SinglePlayerSurround:

  def __init__(self, surround, player_name, opponent, turn_based):

    self._surround = surround
    self._opponent = opponent
    self._player_name = player_name
    self._opponent_name = "white" if player_name is "black" else "black"
    self._turn_based = turn_based

    self._init_state = None
    self._prev_state = None

  def init_state(self):
    """Return init state for game"""

    if self._init_state is None:
      if self._player_name == "black" and self._turn_based:
        state = self._surround.vis_board("white")
        state_reward, END_OF_GAME = self._surround.apply_actions({
          self._opponent_name : self._opponent.next_action(state)
        })
        # assumes opponent can't fail on first move... True as long as doesn't start
        # on border...
        assert not END_OF_GAME and list(state_reward.keys()) == ["black"] and state_reward["black"][1] == 0
      self._init_state = np.copy(self._surround.vis_board(self._player_name))

    return self._init_state

  def apply_action(self, move):
    """
    Args:
      move: oneof ['left', 'right', 'up', 'down']

    Returns:
      tuple (state, reward, end_of_game) with environment response
        state      : next state
        reward     : reward from environment
        end_of_game: Indicates if the player can make any more moves 
    """
    # if turn_based: first apply action of 'player' then opponent
    if self._turn_based:
      state_rewards, END_OF_GAME = self._surround.apply_actions({
        self._player_name : move
      })
      if not END_OF_GAME:
        state, reward = state_rewards[self._opponent_name]
        state_rewards, END_OF_GAME = self._surround.apply_actions({
          self._opponent_name : self._opponent.next_action(state)
        })


    else:
      prev_state = self._prev_state if self._prev_state is not None else self._surround.vis_board(self._opponent_name)
      state_rewards, END_OF_GAME = self._surround.apply_actions({
        self._player_name : move,
        self._opponent_name : self._opponent.next_action(prev_state)
      })
      self._prev_state, _ = state_rewards[self._opponent_name]
    state, reward = state_rewards[self._player_name]

    return state, reward, END_OF_GAME

class Environment:
  """Create environment for agent playing against single opponent. The opponent
  will be fixed during the duration of this environment
  """

  # opponent needs to be shared for batching...
  def __init__(self, height, width, opponent, turn_based=False):
    """
    Args:
      turn_based : if true, each player takes turn playing a move, otherwise
                   both players places their move at the same time
      opponent: agent object which has method 'next_action'
    """
    self.opponent = opponent
    self.height = height
    self.width = width
    self.turn_based = turn_based

  # if turn_based, should specify 'white' or 'black'
  # game or games?
  def new_game(self, player=None):
    """Create new game for player
    
    Args:
      player     : one of ['white', 'black'], only really matters if 'turn_based'
                   as 'white' then moves first
    Returns:
      game object with functions
        init_state   : agent can use to get initial state
        apply_action : agent uses this to apply action
    """
    if player is None:
      player = random.choice(["white", "black"])

    game = Surround(self.height, self.width)
    single_player_game = SinglePlayerSurround(game, player, self.opponent, self.turn_based)

    return single_player_game
